from clickhouse_connect import driver_name
from clickhouse_connect.cc_sqlalchemy.datatypes.base import schema_types
from clickhouse_connect.cc_sqlalchemy.sql import final, sample
from clickhouse_connect.cc_sqlalchemy.sql.clauses import array_join, ArrayJoin, ch_join, ClickHouseJoin

# pylint: disable=invalid-name
dialect_name = driver_name
ischema_names = schema_types

__all__ = ['dialect_name', 'ischema_names', 'array_join', 'ArrayJoin', 'ch_join', 'ClickHouseJoin', 'final', 'sample']



import logging
import os

import clickhouse_connect.driver.dataconv as pydc
import clickhouse_connect.driver.npconv as pync
from clickhouse_connect.driver.buffer import ResponseBuffer
from clickhouse_connect.driver.common import coerce_bool

logger = logging.getLogger(__name__)

RespBuffCls = ResponseBuffer
data_conv = pydc
# numpy_conv is resolved lazily via __getattr__ to avoid eagerly importing numpy


# pylint: disable=import-outside-toplevel,global-statement

def connect_c_modules():
    if not coerce_bool(os.environ.get('CLICKHOUSE_CONNECT_USE_C', True)):
        logger.info('ClickHouse Connect C optimizations disabled')
        return

    global RespBuffCls, data_conv
    try:
        from clickhouse_connect.driverc.buffer import ResponseBuffer as CResponseBuffer
        import clickhouse_connect.driverc.dataconv as cdc

        data_conv = cdc
        RespBuffCls = CResponseBuffer
        logger.debug('Successfully imported ClickHouse Connect C data optimizations')
    except ImportError as ex:
        logger.warning('Unable to connect optimized C data functions [%s], falling back to pure Python',
                       str(ex))


def _resolve_numpy_conv():
    if "numpy_conv" in globals():
        return
    if coerce_bool(os.environ.get('CLICKHOUSE_CONNECT_USE_C', True)):
        try:
            import clickhouse_connect.driverc.npconv as cnc
            globals()["numpy_conv"] = cnc
            logger.debug('Successfully import ClickHouse Connect C/Numpy optimizations')
            return
        except ImportError as ex:
            logger.debug('Unable to connect ClickHouse Connect C to Numpy API [%s], falling back to pure Python',
                         str(ex))
    globals()["numpy_conv"] = pync


def __getattr__(name):
    if name == "numpy_conv":
        _resolve_numpy_conv()
        return globals()["numpy_conv"]
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


connect_c_modules()



import logging
import re
import warnings
import pytz

from io import IOBase
from typing import Any, Literal, Tuple, Dict, Sequence, Optional, Union, Generator, BinaryIO, TYPE_CHECKING
from datetime import tzinfo

from pytz.exceptions import UnknownTimeZoneError

from clickhouse_connect.driver import tzutil
from clickhouse_connect.driver.binding import bind_query
from clickhouse_connect.driver.common import dict_copy, empty_gen, StreamContext, get_rename_method
from clickhouse_connect.driver.external import ExternalData
from clickhouse_connect.driver.types import Matrix, Closable
from clickhouse_connect.driver.exceptions import StreamClosedError, ProgrammingError
from clickhouse_connect.driver import options
from clickhouse_connect.driver.options import check_arrow
from clickhouse_connect.driver.context import BaseQueryContext

if TYPE_CHECKING:
    from clickhouse_connect.datatypes.base import ClickHouseType

logger = logging.getLogger(__name__)

TzMode = Literal["naive_utc", "aware", "schema"]
TzSource = Literal["auto", "server", "local"]

_UTC_TZ_AWARE_TO_TZ_MODE: Dict[Union[bool, str], TzMode] = {
    False: "naive_utc",
    True: "aware",
    "schema": "schema",
}

_VALID_TZ_MODES = {"naive_utc", "aware", "schema"}

_TZ_MODE_TO_UTC_TZ_AWARE: Dict[str, Union[bool, Literal["schema"]]] = {
    "naive_utc": False,
    "aware": True,
    "schema": "schema",
}

_APPLY_SERVER_TZ_TO_TZ_SOURCE: Dict[Union[bool, str, None], TzSource] = {
    None: "auto",
    True: "server",
    False: "local",
    "always": "server",
}

_TZ_SOURCE_TO_APPLY_SERVER_TZ: Dict[str, Optional[Union[bool, str]]] = {
    "auto": None,
    "server": True,
    "local": False,
}

_VALID_TZ_SOURCES = {"auto", "server", "local"}

# Mapping for string booleans that may arrive via URL params
_STR_BOOL_MAP = {"true": True, "false": False, "1": True, "0": False}


def _resolve_tz_mode(
    tz_mode: Optional[TzMode] = None,
    utc_tz_aware: Optional[Union[bool, Literal["schema"]]] = None,
) -> TzMode:
    """Resolve tz_mode from either the new ``tz_mode`` or deprecated ``utc_tz_aware`` parameter.

    Returns the canonical TzMode string.  Raises ``ProgrammingError`` on conflicts or
    invalid values.
    """
    if tz_mode is not None and utc_tz_aware is not None:
        raise ProgrammingError(
            "Cannot specify both 'tz_mode' and 'utc_tz_aware'. "
            "Use 'tz_mode' only; 'utc_tz_aware' is deprecated."
        )

    if utc_tz_aware is not None:
        # Coerce string booleans from URL params (e.g. "true" -> True)
        if isinstance(utc_tz_aware, str) and utc_tz_aware.lower() in _STR_BOOL_MAP:
            utc_tz_aware = _STR_BOOL_MAP[utc_tz_aware.lower()]

        if utc_tz_aware not in _UTC_TZ_AWARE_TO_TZ_MODE:
            raise ProgrammingError(
                f'utc_tz_aware must be True, False, or "schema", got "{utc_tz_aware}"'
            )
        warnings.warn(
            "utc_tz_aware is deprecated and will be removed in 1.0. "
            "Use tz_mode='naive_utc' | 'aware' | 'schema' instead.",
            DeprecationWarning,
            stacklevel=3,
        )
        return _UTC_TZ_AWARE_TO_TZ_MODE[utc_tz_aware]

    if tz_mode is not None:
        if tz_mode not in _VALID_TZ_MODES:
            raise ProgrammingError(
                f'tz_mode must be "naive_utc", "aware", or "schema", got "{tz_mode}"'
            )
        return tz_mode

    return "naive_utc"


def _resolve_tz_source(
    tz_source: Optional[TzSource] = None,
    apply_server_timezone: Optional[Union[str, bool]] = None,
) -> TzSource:
    """Resolve tz_source from either the new ``tz_source`` or deprecated ``apply_server_timezone`` parameter.

    Returns the canonical TzSource string.  Raises ``ProgrammingError`` on conflicts or
    invalid values.
    """
    if tz_source is not None and apply_server_timezone is not None:
        raise ProgrammingError(
            "Cannot specify both 'tz_source' and 'apply_server_timezone'. "
            "Use 'tz_source' only; 'apply_server_timezone' is deprecated."
        )

    if apply_server_timezone is not None:
        # Coerce string booleans from URL params (e.g. "true" -> True)
        if isinstance(apply_server_timezone, str) and apply_server_timezone.lower() in _STR_BOOL_MAP:
            apply_server_timezone = _STR_BOOL_MAP[apply_server_timezone.lower()]

        if apply_server_timezone not in _APPLY_SERVER_TZ_TO_TZ_SOURCE:
            raise ProgrammingError(
                f"apply_server_timezone must be None, True, False, or 'always', "
                f'got "{apply_server_timezone}"'
            )
        warnings.warn(
            "apply_server_timezone is deprecated and will be removed in 1.0. "
            "Use tz_source='auto' | 'server' | 'local' instead.",
            DeprecationWarning,
            stacklevel=3,
        )
        return _APPLY_SERVER_TZ_TO_TZ_SOURCE[apply_server_timezone]

    if tz_source is not None:
        if tz_source not in _VALID_TZ_SOURCES:
            raise ProgrammingError(
                f'tz_source must be "auto", "server", or "local", got "{tz_source}"'
            )
        return tz_source

    return "auto"


commands = 'CREATE|ALTER|SYSTEM|GRANT|REVOKE|CHECK|DETACH|ATTACH|DROP|DELETE|KILL|' + \
           'OPTIMIZE|SET|RENAME|TRUNCATE|USE|UPDATE'

limit_re = re.compile(r'\s+LIMIT($|\s)', re.IGNORECASE)
select_re = re.compile(r'(^|\s)SELECT\s', re.IGNORECASE)
insert_re = re.compile(r'(^|\s)INSERT\s*INTO', re.IGNORECASE)
command_re = re.compile(r'(^\s*)(' + commands + r')\s', re.IGNORECASE)


# pylint: disable=too-many-instance-attributes
class QueryContext(BaseQueryContext):
    """
    Argument/parameter object for queries.  This context is used to set thread/query specific formats
    """

    # pylint: disable=duplicate-code,too-many-arguments,too-many-positional-arguments,too-many-locals
    def __init__(self,
                 query: Union[str, bytes] = '',
                 parameters: Optional[Dict[str, Any]] = None,
                 settings: Optional[Dict[str, Any]] = None,
                 query_formats: Optional[Dict[str, str]] = None,
                 column_formats: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
                 encoding: Optional[str] = None,
                 server_tz: tzinfo = pytz.UTC,
                 use_none: Optional[bool] = None,
                 column_oriented: Optional[bool] = None,
                 use_numpy: Optional[bool] = None,
                 max_str_len: Optional[int] = 0,
                 query_tz: Optional[Union[str, tzinfo]] = None,
                 column_tzs: Optional[Dict[str, Union[str, tzinfo]]] = None,
                 utc_tz_aware: Optional[Union[bool, Literal["schema"]]] = None,
                 use_extended_dtypes: Optional[bool] = None,
                 as_pandas: bool = False,
                 streaming: bool = False,
                 apply_server_tz: bool = False,
                 external_data: Optional[ExternalData] = None,
                 transport_settings: Optional[Dict[str, str]] = None,
                 rename_response_column: Optional[str] = None,
                 tz_mode: Optional[TzMode] = None):
        """
        Initializes various configuration settings for the query context

        :param query:  Query string with Python style format value replacements
        :param parameters: Optional dictionary of substitution values
        :param settings: Optional ClickHouse settings for the query
        :param query_formats: Optional dictionary of query formats with the key of a ClickHouse type name
          (with * wildcards) and a value of valid query formats for those types.
          The value 'encoding' can be sent to change the expected encoding for this query, with a value of
          the desired encoding such as `latin-1`
        :param column_formats: Optional dictionary of column specific formats.  The key is the column name,
          The value is either the format for the data column (such as 'string' for a UUID column) or a
          second level "format" dictionary of a ClickHouse type name and a value of query formats.  This
          secondary dictionary can be used for nested column types such as Tuples or Maps
        :param encoding: Optional string encoding for this query, such as 'latin-1'
        :param column_formats: Optional dictionary
        :param use_none: Use a Python None for ClickHouse NULL values in nullable columns.  Otherwise the default
          value of the column (such as 0 for numbers) will be returned in the result_set
        :param max_str_len Limit returned ClickHouse String values to this length, which allows a Numpy
          structured array even with ClickHouse variable length String columns.  If 0, Numpy arrays for
          String columns will always be object arrays
        :param query_tz  Either a string or a pytz tzinfo object.  (Strings will be converted to tzinfo objects).
          Values for any DateTime or DateTime64 column in the query will be converted to Python datetime.datetime
          objects with the selected timezone
        :param column_tzs A dictionary of column names to tzinfo objects (or strings that will be converted to
          tzinfo objects).  The timezone will be applied to datetime objects returned in the query
        :param tz_mode Controls timezone-aware behavior for UTC DateTime columns. "naive_utc" (default) returns
          naive UTC timestamps. "aware" forces timezone-aware UTC datetimes. "schema" returns datetimes that
          match the server's column definition which means timezone-aware when the column schema defines a timezone
          (e.g. DateTime('UTC')) and naive for bare DateTime columns.
        :param utc_tz_aware Deprecated. Use tz_mode instead.
        """
        super().__init__(settings,
                         query_formats,
                         column_formats,
                         encoding,
                         use_extended_dtypes if use_extended_dtypes is not None else False,
                         use_numpy if use_numpy is not None else False,
                         transport_settings=transport_settings)
        self.query = query
        self.parameters = parameters or {}
        self.use_none = True if use_none is None else use_none
        self.column_oriented = False if column_oriented is None else column_oriented
        self.use_numpy = use_numpy
        self.max_str_len = 0 if max_str_len is None else max_str_len
        self.server_tz = server_tz
        self.apply_server_tz = apply_server_tz
        self.external_data = external_data
        self.tz_mode = _resolve_tz_mode(tz_mode, utc_tz_aware)
        if isinstance(query_tz, str):
            try:
                query_tz = pytz.timezone(query_tz)
            except UnknownTimeZoneError as ex:
                raise ProgrammingError(f'query_tz {query_tz} is not recognized') from ex
        self.query_tz = query_tz
        if column_tzs is not None:
            for col_name, timezone in column_tzs.items():
                if isinstance(timezone, str):
                    try:
                        timezone = pytz.timezone(timezone)
                        column_tzs[col_name] = timezone
                    except UnknownTimeZoneError as ex:
                        raise ProgrammingError(f'column_tz {timezone} is not recognized') from ex
        self.column_tzs = column_tzs
        self.column_tz = None
        self.response_tz = None
        self.block_info = False
        self.as_pandas = as_pandas
        self.use_pandas_na = as_pandas and options.pd_extended_dtypes
        self.streaming = streaming
        self._rename_response_column: Optional[str] = rename_response_column
        self.column_renamer = get_rename_method(rename_response_column)
        self._update_query()

    @property
    def rename_response_column(self) -> Optional[str]:
        return self._rename_response_column

    @rename_response_column.setter
    def rename_response_column(self, method: Optional[str]):
        self._rename_response_column = method
        self.column_renamer = get_rename_method(method)

    @property
    def is_select(self) -> bool:
        return select_re.search(self.uncommented_query) is not None

    @property
    def has_limit(self) -> bool:
        return limit_re.search(self.uncommented_query) is not None

    @property
    def is_insert(self) -> bool:
        return insert_re.search(self.uncommented_query) is not None

    @property
    def is_command(self) -> bool:
        return command_re.search(self.uncommented_query) is not None

    @property
    def utc_tz_aware(self) -> Union[bool, Literal["schema"]]:
        """Deprecated: use tz_mode instead."""
        warnings.warn(
            "utc_tz_aware is deprecated and will be removed in 1.0. "
            "Use tz_mode instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return _TZ_MODE_TO_UTC_TZ_AWARE[self.tz_mode]

    def set_parameters(self, parameters: Dict[str, Any]):
        self.parameters = parameters
        self._update_query()

    def set_parameter(self, key: str, value: Any):
        if not self.parameters:
            self.parameters = {}
        self.parameters[key] = value
        self._update_query()

    def set_response_tz(self, response_tz: tzinfo):
        self.response_tz = response_tz

    def start_column(self, name: str):
        super().start_column(name)
        if self.column_tzs and name in self.column_tzs:
            self.column_tz = self.column_tzs[name]
        else:
            self.column_tz = None

    def active_tz(self, datatype_tz: Optional[tzinfo]):
        if self.tz_mode == "schema":
            return self.column_tz or datatype_tz
        if self.column_tz:
            active_tz = self.column_tz
        elif datatype_tz:
            active_tz = datatype_tz
        elif self.query_tz:
            active_tz = self.query_tz
        elif self.response_tz:
            active_tz = self.response_tz
        elif self.apply_server_tz:
            active_tz = self.server_tz
        else:
            active_tz = tzutil.local_tz
        if tzutil.is_utc_timezone(active_tz) and self.tz_mode == "naive_utc":
            return None
        return active_tz

    # pylint disable=too-many-positional-arguments
    def updated_copy(self,
                     query: Optional[Union[str, bytes]] = None,
                     parameters: Optional[Dict[str, Any]] = None,
                     settings: Optional[Dict[str, Any]] = None,
                     query_formats: Optional[Dict[str, str]] = None,
                     column_formats: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
                     encoding: Optional[str] = None,
                     server_tz: Optional[tzinfo] = None,
                     use_none: Optional[bool] = None,
                     column_oriented: Optional[bool] = None,
                     use_numpy: Optional[bool] = None,
                     max_str_len: Optional[int] = None,
                     query_tz: Optional[Union[str, tzinfo]] = None,
                     column_tzs: Optional[Dict[str, Union[str, tzinfo]]] = None,
                     utc_tz_aware: Optional[Union[bool, Literal["schema"]]] = None,
                     use_extended_dtypes: Optional[bool] = None,
                     as_pandas: bool = False,
                     streaming: bool = False,
                     external_data: Optional[ExternalData] = None,
                     transport_settings: Optional[Dict[str, str]] = None,
                     rename_response_column: Optional[str] = None,
                     tz_mode: Optional[TzMode] = None) -> 'QueryContext':
        """
        Creates Query context copy with parameters overridden/updated as appropriate.
        """
        if tz_mode is not None or utc_tz_aware is not None:
            resolved_tz_mode = _resolve_tz_mode(tz_mode, utc_tz_aware)
        else:
            resolved_tz_mode = self.tz_mode
        return QueryContext(
            query=query or self.query,
            parameters=dict_copy(self.parameters, parameters),
            settings=dict_copy(self.settings, settings),
            query_formats=dict_copy(self.query_formats, query_formats),
            column_formats=dict_copy(self.column_formats, column_formats),
            encoding=encoding if encoding else self.encoding,
            server_tz=server_tz if server_tz else self.server_tz,
            use_none=self.use_none if use_none is None else use_none,
            column_oriented=self.column_oriented if column_oriented is None else column_oriented,
            use_numpy=self.use_numpy if use_numpy is None else use_numpy,
            max_str_len=self.max_str_len if max_str_len is None else max_str_len,
            query_tz=self.query_tz if query_tz is None else query_tz,
            column_tzs=self.column_tzs if column_tzs is None else column_tzs,
            tz_mode=resolved_tz_mode,
            use_extended_dtypes=self.use_extended_dtypes if use_extended_dtypes is None else use_extended_dtypes,
            as_pandas=as_pandas,
            streaming=streaming,
            apply_server_tz=self.apply_server_tz,
            external_data=self.external_data if external_data is None else external_data,
            transport_settings=self.transport_settings if transport_settings is None else transport_settings,
            rename_response_column=self.rename_response_column if rename_response_column is None else rename_response_column,
        )

    def _update_query(self):
        self.final_query, self.bind_params = bind_query(self.query, self.parameters, self.server_tz)
        if isinstance(self.final_query, bytes):
            # If we've embedded binary data in the query, all bets are off, and we check the original query for comments
            self.uncommented_query = remove_sql_comments(self.query)
        else:
            self.uncommented_query = remove_sql_comments(self.final_query)


class QueryResult(Closable):
    """
    Wrapper class for query return values and metadata
    """

    # pylint: disable=too-many-arguments
    def __init__(self,
                 result_set: Matrix = None,
                 block_gen: Generator[Matrix, None, None] = None,
                 column_names: Tuple[str, ...] = (),
                 column_types: Tuple['ClickHouseType', ...] = (),
                 column_oriented: bool = False,
                 source: Closable = None,
                 query_id: str = None,
                 summary: Dict[str, Any] = None):
        self._result_rows = result_set
        self._result_columns = None
        self._block_gen = block_gen or empty_gen()
        self._in_context = False
        self._query_id = query_id
        self.column_names = column_names
        self.column_types = column_types
        self.column_oriented = column_oriented
        self.source = source
        self.summary = {} if summary is None else summary

    @property
    def result_set(self) -> Matrix:
        if self.column_oriented:
            return self.result_columns
        return self.result_rows

    @property
    def result_columns(self) -> Matrix:
        if self._result_columns is None:
            result = [[] for _ in range(len(self.column_names))]
            with self.column_block_stream as stream:
                for block in stream:
                    for base, added in zip(result, block):
                        base.extend(added)
            self._result_columns = result
        return self._result_columns

    @property
    def result_rows(self) -> Matrix:
        if self._result_rows is None:
            result = []
            with self.row_block_stream as stream:
                for block in stream:
                    result.extend(block)
            self._result_rows = result
        return self._result_rows

    @property
    def query_id(self) -> str:
        query_id = self.summary.get('query_id')
        if query_id:
            return query_id
        return self._query_id

    def _column_block_stream(self):
        if self._block_gen is None:
            raise StreamClosedError
        block_stream = self._block_gen
        self._block_gen = None
        return block_stream

    def _row_block_stream(self):
        for block in self._column_block_stream():
            yield list(zip(*block))

    @property
    def column_block_stream(self) -> StreamContext:
        return StreamContext(self, self._column_block_stream())

    @property
    def row_block_stream(self) -> StreamContext:
        return StreamContext(self, self._row_block_stream())

    @property
    def rows_stream(self) -> StreamContext:
        def stream():
            for block in self._row_block_stream():
                yield from block

        return StreamContext(self, stream())

    def named_results(self) -> Generator[dict, None, None]:
        for row in zip(*self.result_set) if self.column_oriented else self.result_set:
            yield dict(zip(self.column_names, row))

    @property
    def row_count(self) -> int:
        if self.column_oriented:
            return 0 if len(self.result_set) == 0 else len(self.result_set[0])
        return len(self.result_set)

    @property
    def first_item(self) -> Dict[str, Any]:
        if self.column_oriented:
            return {name: col[0] for name, col in zip(self.column_names, self.result_set)}
        return dict(zip(self.column_names, self.result_set[0]))

    @property
    def first_row(self) -> Sequence[Any]:
        if self.column_oriented:
            return [col[0] for col in self.result_set]
        return self.result_set[0]

    def close(self) -> None:
        if self.source:
            self.source.close()
            self.source = None
        if self._block_gen is not None:
            self._block_gen.close()
            self._block_gen = None


comment_re = re.compile(r"(\".*?\"|\'.*?\')|(/\*.*?\*/|(--\s)[^\n]*$)", re.MULTILINE | re.DOTALL)


def remove_sql_comments(sql: str) -> str:
    """
    Remove SQL comments.  This is useful to determine the type of SQL query, such as SELECT or INSERT, but we
    don't fully trust it to correctly ignore weird quoted strings, and other edge cases, so we always pass the
    original SQL to ClickHouse (which uses a full-fledged AST/ token parser)
    :param sql:  SQL query
    :return: SQL Query without SQL comments
    """

    def replacer(match):
        # if the 2nd group (capturing comments) is not None, it means we have captured a
        # non-quoted, actual comment string, so return nothing to remove the comment
        if match.group(2):
            return ''
        # Otherwise we've actually captured a quoted string, so return it
        return match.group(1)

    return comment_re.sub(replacer, sql)


def to_arrow(content: bytes):
    pyarrow = check_arrow()
    reader = pyarrow.ipc.RecordBatchFileReader(content)
    return reader.read_all()


def to_arrow_batches(buffer: IOBase) -> StreamContext:
    pyarrow = check_arrow()
    reader = pyarrow.ipc.open_stream(buffer)
    return StreamContext(buffer, reader)


def arrow_buffer(table, compression: Optional[str] = None) -> Tuple[Sequence[str], Union[bytes, BinaryIO]]:
    pyarrow = check_arrow()
    write_options = None
    if compression in ('zstd', 'lz4'):
        write_options = pyarrow.ipc.IpcWriteOptions(compression=pyarrow.Codec(compression=compression))
    sink = pyarrow.BufferOutputStream()
    with pyarrow.RecordBatchFileWriter(sink, table.schema, options=write_options) as writer:
        writer.write(table)
    return table.schema.names, sink.getvalue()



from typing import Optional

from sqlalchemy import and_, true
from sqlalchemy.sql.base import Immutable
from sqlalchemy.sql.selectable import FromClause, Join
from sqlalchemy.sql.visitors import InternalTraversal


def _normalize_array_columns(array_column, alias):
    """Normalize single/multi column input into a list of (column, alias_or_none) tuples."""
    if isinstance(array_column, (list, tuple)):
        columns = list(array_column)
        if not columns:
            raise ValueError("At least one array column is required")
        if alias is None:
            aliases = [None] * len(columns)
        elif isinstance(alias, (list, tuple)):
            aliases = list(alias)
            if len(aliases) != len(columns):
                raise ValueError(f"Length of alias list ({len(aliases)}) must match " f"length of array_column list ({len(columns)})")
        else:
            raise ValueError("alias must be a list when array_column is a list")
    else:
        columns = [array_column]
        if isinstance(alias, (list, tuple)):
            raise ValueError("alias must be a string or None when array_column is a single column")
        aliases = [alias]

    return list(zip(columns, aliases))


# pylint: disable=protected-access,too-many-ancestors,abstract-method,unused-argument
class ArrayJoin(Immutable, FromClause):
    """Represents ClickHouse ARRAY JOIN clause.

    Supports single or multiple array columns with optional per-column aliases.
    Multiple columns are expanded in parallel (zipped by position), not as a
    cartesian product. All arrays in a single ARRAY JOIN must have the same
    length per row unless enable_unaligned_array_join is set on the server.

    See: https://clickhouse.com/docs/sql-reference/statements/select/array-join
    """

    __visit_name__ = "array_join"
    _is_from_container = True
    named_with_column = False
    _is_join = True

    def __init__(self, left, array_column, alias=None, is_left=False):
        """Initialize ARRAY JOIN clause.

        Args:
            left: The left side (table or subquery).
            array_column: A single array column, or a list/tuple of array columns.
            alias: Optional alias. A single string when array_column is a single
                column, or a list/tuple of strings (same length as array_column)
                when array_column is a list. None means no aliases.
            is_left: If True, use LEFT ARRAY JOIN instead of ARRAY JOIN.
        """
        super().__init__()
        self.left = left
        self.array_columns = _normalize_array_columns(array_column, alias)
        self.is_left = is_left
        self._is_clone_of = None

    @property
    def selectable(self):
        """Return the selectable for this clause"""
        return self.left

    @property
    def _hide_froms(self):
        """Hide the left table from the FROM clause since it's part of the ARRAY JOIN"""
        return [self.left]

    @property
    def _from_objects(self):
        """Return all FROM objects referenced by this construct"""
        return self.left._from_objects

    def _clone(self, **kw):
        """Return a copy of this ArrayJoin"""
        c = self.__class__.__new__(self.__class__)
        c.__dict__ = self.__dict__.copy()
        c._is_clone_of = self
        return c

    def _copy_internals(self, clone=None, **kw):
        """Copy internal state for cloning.

        This ensures that when queries are cloned (e.g., for subqueries, unions, or CTEs),
        the left FromClause and array column references are properly deep-cloned.
        """
        def _default_clone(elem, **kwargs):
            return elem

        if clone is None:
            clone = _default_clone

        self.left = clone(self.left, **kw)
        self.array_columns = [
            (clone(col, **kw), alias)
            for col, alias in self.array_columns
        ]


def array_join(left, array_column, alias=None, is_left=False):
    """Create an ARRAY JOIN clause.

    Supports single or multiple array columns. When multiple columns are
    provided, they are expanded in parallel (zipped by index position).

    Args:
        left: The left side (table or subquery).
        array_column: A single array column, or a list/tuple of array columns.
        alias: Optional alias. A single string when array_column is a single
            column, or a list/tuple of strings (same length as array_column)
            when array_column is a list. None means no aliases.
        is_left: If True, use LEFT ARRAY JOIN instead of ARRAY JOIN.

    Returns:
        ArrayJoin: An ArrayJoin clause element.

    Examples:
        from clickhouse_connect.cc_sqlalchemy.sql.clauses import array_join

        # Single column ARRAY JOIN
        query = select(table).select_from(array_join(table, table.c.tags))

        # Single column LEFT ARRAY JOIN with alias
        query = select(table).select_from(
            array_join(table, table.c.tags, alias="tag", is_left=True)
        )

        # Multiple columns with aliases
        query = select(table).select_from(
            array_join(
                table,
                [table.c.names, table.c.prices, table.c.quantities],
                alias=["name", "price", "quantity"],
            )
        )
    """
    return ArrayJoin(left, array_column, alias, is_left)


_VALID_STRICTNESS = frozenset({None, "ALL", "ANY", "SEMI", "ANTI", "ASOF"})
_VALID_DISTRIBUTION = frozenset({None, "GLOBAL"})


def _validate_ch_join(strictness, distribution, onclause, isouter, full, is_cross, using):
    """Validate ClickHouse join parameter combinations."""
    if strictness not in _VALID_STRICTNESS:
        raise ValueError(f"Invalid strictness {strictness!r}. Must be one of: ALL, ANY, SEMI, ANTI, ASOF")
    if distribution not in _VALID_DISTRIBUTION:
        raise ValueError(f"Invalid distribution {distribution!r}. Must be: GLOBAL")
    if is_cross and strictness is not None:
        raise ValueError("Strictness modifiers cannot be used with CROSS JOIN")
    if is_cross and (isouter or full):
        raise ValueError("CROSS JOIN cannot be combined with isouter or full")
    if strictness in ("SEMI", "ANTI") and not isouter:
        raise ValueError(f"{strictness} JOIN requires isouter=True (LEFT) or swapped table order (RIGHT)")
    if strictness == "ASOF" and full:
        raise ValueError("ASOF is not supported with FULL joins")
    if using is not None:
        if is_cross:
            raise ValueError("USING cannot be combined with CROSS JOIN")
        if onclause is not None:
            raise ValueError("Cannot specify both onclause and using")
        if not isinstance(using, (list, tuple)) or not using:
            raise ValueError("using must be a non-empty list of column name strings")
        if not all(isinstance(col, str) for col in using):
            raise ValueError("using must contain only column name strings")


def _build_using_onclause(left, right, using):
    """Build an equality onclause from USING column names.

    This gives SQLAlchemy's from-linter proper column references so it
    knows the tables are connected. The compiler renders USING instead of ON.
    """
    conditions = []
    for col in using:
        try:
            conditions.append(left.c[col] == right.c[col])
        except KeyError:
            left_cols = {c.name for c in left.c}
            right_cols = {c.name for c in right.c}
            missing_from = []
            if col not in left_cols:
                missing_from.append(str(left))
            if col not in right_cols:
                missing_from.append(str(right))
            raise ValueError(f"USING column {col!r} not found in: {', '.join(missing_from)}") from None
    return and_(*conditions) if len(conditions) > 1 else conditions[0]


# pylint: disable=too-many-ancestors,abstract-method
class ClickHouseJoin(Join):
    """A SQLAlchemy Join subclass that supports ClickHouse-specific join features.

    ClickHouse JOIN syntax: [GLOBAL] [ALL|ANY|SEMI|ANTI|ASOF] [INNER|LEFT|RIGHT|FULL|CROSS] JOIN

    Strictness modifiers control how multiple matches are handled:
        - ALL: return all matching rows (default, standard SQL behavior)
        - ANY: return only the first match per left row
        - SEMI: acts as an allowlist on join keys, no Cartesian product
        - ANTI: acts as a denylist on join keys, no Cartesian product
        - ASOF: time-series join, finds the closest match

    Distribution modifier:
        - GLOBAL: broadcasts the right table to all nodes in distributed queries

    USING clause:
        - Joins on same-named columns from both tables. Unlike ON, USING merges
          matched columns into one, which is important for FULL OUTER JOIN where
          ON produces default values (0, '') for unmatched sides.

    Note: RIGHT JOIN is achieved by swapping table order, which is standard SQLAlchemy behavior.
    ASOF JOIN requires the last ON condition to be an inequality which is validated by
    the ClickHouse server, not here. Not all strictness/join type combinations are supported
    by every join algorithm and the server will report unsupported combinations.
    """

    __visit_name__ = "join"

    _traverse_internals = Join._traverse_internals + [
        ("strictness", InternalTraversal.dp_string),
        ("distribution", InternalTraversal.dp_string),
        ("_is_cross", InternalTraversal.dp_boolean),
        ("using_columns", InternalTraversal.dp_string_list),
    ]

    def __init__(self, left, right, onclause=None, isouter=False, full=False,
                 strictness=None, distribution=None, _is_cross=False, using=None):
        if strictness is not None:
            strictness = strictness.upper()
        if distribution is not None:
            distribution = distribution.upper()

        _validate_ch_join(strictness, distribution, onclause, isouter, full, _is_cross, using)

        effective_onclause = _build_using_onclause(left, right, using) if using else onclause
        super().__init__(left, right, effective_onclause, isouter, full)
        self.strictness = strictness
        self.distribution = distribution
        self._is_cross = _is_cross
        self.using_columns = list(using) if using is not None else None


def ch_join(
    left,
    right,
    onclause=None,
    *,
    isouter=False,
    full=False,
    cross=False,
    using=None,
    strictness: Optional[str] = None,
    distribution: Optional[str] = None,
):
    """Create a ClickHouse JOIN with optional strictness, distribution, and USING support.

    Args:
        left: The left side table or selectable.
        right: The right side table or selectable.
        onclause: The ON clause expression. Mutually exclusive with ``using``.
        isouter: If True, render a LEFT OUTER JOIN.
        full: If True, render a FULL OUTER JOIN.
        cross: If True, render a CROSS JOIN. Cannot be combined with
            onclause, using, or strictness modifiers.
        using: A list of column name strings for USING syntax. The columns
            must have the same name in both tables. Mutually exclusive with
            ``onclause``. Produces ``USING (col1, col2)`` instead of ``ON``.
        strictness: ClickHouse strictness modifier, one of
            "ALL", "ANY", "SEMI", "ANTI", or "ASOF".
        distribution: ClickHouse distribution modifier "GLOBAL".

    Returns:
        ClickHouseJoin: A join element with ClickHouse modifiers.
    """
    if cross:
        if onclause is not None:
            raise ValueError("cross=True conflicts with an explicit onclause")
        if using is not None:
            raise ValueError("cross=True conflicts with using")
        onclause = true()
    return ClickHouseJoin(left, right, onclause, isouter, full,
                          strictness, distribution, _is_cross=cross, using=using)



from typing import NamedTuple, Sequence

from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.driver.options import check_arrow


class TableColumnDef(NamedTuple):
    """
    Simplified ClickHouse Table Column definition for DDL
    """
    name: str
    ch_type: ClickHouseType
    expr_type: str = None
    expr: str = None

    @property
    def col_expr(self):
        expr = f'{self.name} {self.ch_type.name}'
        if self.expr_type:
            expr += f' {self.expr_type} {self.expr}'
        return expr


def create_table(table_name: str, columns: Sequence[TableColumnDef], engine: str, engine_params: dict):
    stmt = f"CREATE TABLE {table_name} ({', '.join(col.col_expr for col in columns)}) ENGINE {engine} "
    if engine_params:
        for key, value in engine_params.items():
            stmt += f' {key} {value}'
    return stmt


def _arrow_type_to_ch(arrow_type: "pa.DataType") -> str: # pylint: disable=too-many-return-statements,too-many-branches
    """
    Best-effort mapping from common PyArrow types to ClickHouse type names.

    Covers core scalar and common date/time/timestamp types. For anything unknown, we raise, so the
    caller is aware that the automatic mapping is not implemented for that Arrow type.
    """
    pa = check_arrow()

    pat = pa.types

    # Signed ints
    if pat.is_int8(arrow_type):
        return 'Int8'
    if pat.is_int16(arrow_type):
        return 'Int16'
    if pat.is_int32(arrow_type):
        return 'Int32'
    if pat.is_int64(arrow_type):
        return 'Int64'

    # Unsigned ints
    if pat.is_uint8(arrow_type):
        return 'UInt8'
    if pat.is_uint16(arrow_type):
        return 'UInt16'
    if pat.is_uint32(arrow_type):
        return 'UInt32'
    if pat.is_uint64(arrow_type):
        return 'UInt64'

    # Floats
    if pat.is_float16(arrow_type) or pat.is_float32(arrow_type):
        return 'Float32'
    if pat.is_float64(arrow_type):
        return 'Float64'

    # Boolean
    if pat.is_boolean(arrow_type):
        return 'Bool'

    # Dates
    if pat.is_date32(arrow_type):
        return 'Date32'
    if pat.is_date64(arrow_type):
        return 'DateTime64(3)'

    # Timestamps → DateTime / DateTime64
    if pat.is_timestamp(arrow_type):
        unit = getattr(arrow_type, 'unit', 's')
        tz = getattr(arrow_type, 'tz', None)

        if unit == 's':
            base = 'DateTime'
            if tz:
                return f"DateTime('{tz}')"
            return base

        scale_map = {'ms': 3, 'us': 6, 'ns': 9}
        scale = scale_map.get(unit, 3)
        if tz:
            return f"DateTime64({scale}, '{tz}')"
        return f'DateTime64({scale})'

    # Strings (this covers pa.string(), pa.large_string())
    if pat.is_string(arrow_type) or pat.is_large_string(arrow_type):
        return 'String'

    # for any currently unsupported type, we raise so it’s clear that
    # this Arrow type isn’t supported by the helper yet.
    raise TypeError(f'Unsupported Arrow type for automatic mapping: {arrow_type!r}')


class _DDLType:
    """
    Minimal helper used to satisfy TableColumnDef.ch_type.

    create_table() only needs ch_type.name when building the DDL string,
    so we'll wrap the ClickHouse type name in this tiny object instead of
    constructing full ClickHouseType instances here.
    """
    def __init__(self, name: str):
        self.name = name


def arrow_schema_to_column_defs(schema: "pa.Schema") -> list[TableColumnDef]:
    """
    Convert a PyArrow Schema into a list of TableColumnDef objects.
    
    This helper uses an *optimistic non-null* strategy: it always produces
    non-nullable ClickHouse types, even though Arrow fields are nullable by
    default.

    Note that if the user inserts a table with nulls into a non-Nullable column,
    ClickHouse will silently convert those nulls to default values due to the default
    server setting input_format_null_as_default=1 and current lack of client-side
    validation on arrow inserts.
    """
    pa = check_arrow()

    if not isinstance(schema, pa.Schema):
        raise TypeError(f'Expected pyarrow.Schema, got {type(schema)!r}')

    col_defs: list[TableColumnDef] = []
    for field in schema:
        ch_type_name = _arrow_type_to_ch(field.type)
        col_defs.append(
            TableColumnDef(
                name=field.name,
                ch_type=_DDLType(ch_type_name),
            )
        )
    return col_defs


def create_table_from_arrow_schema(
    table_name: str,
    schema: "pa.Schema",
    engine: str,
    engine_params: dict,
) -> str:
    """
    Helper function to build a CREATE TABLE statement from a PyArrow Schema.

    Internally:
      schema -> arrow_schema_to_column_defs -> create_table(...)
    """
    col_defs = arrow_schema_to_column_defs(schema)
    return create_table(table_name, col_defs, engine, engine_params)



import asyncio
import warnings
from concurrent.futures import ThreadPoolExecutor
from inspect import signature
from typing import Optional, Union, Dict, Any
from urllib.parse import urlparse, parse_qs

import clickhouse_connect.driver.ctypes
from clickhouse_connect.driver.client import Client
from clickhouse_connect.driver.common import dict_copy
from clickhouse_connect.driver.exceptions import ProgrammingError
from clickhouse_connect.driver.httpclient import HttpClient
from clickhouse_connect.driver.asyncclient import AsyncClient, DefaultThreadPoolExecutor, NEW_THREAD_POOL_EXECUTOR

__all__ = ['Client', 'AsyncClient', 'create_client', 'create_async_client']


# pylint: disable=too-many-arguments,too-many-locals,too-many-branches
def create_client(*,
                  host: Optional[str] = None,
                  username: Optional[str] = None,
                  password: str = '',
                  access_token: Optional[str] = None,
                  database: str = '__default__',
                  interface: Optional[str] = None,
                  port: int = 0,
                  secure: Union[bool, str] = False,
                  dsn: Optional[str] = None,
                  settings: Optional[Dict[str, Any]] = None,
                  generic_args: Optional[Dict[str, Any]] = None,
                  **kwargs) -> Client:
    """
    The preferred method to get a ClickHouse Connect Client instance

    :param host: The hostname or IP address of the ClickHouse server. If not set, localhost will be used.
    :param username: The ClickHouse username. If not set, the default ClickHouse user will be used.
      Should not be set if `access_token` is used.
    :param password: The password for username.
      Should not be set if `access_token` is used.
    :param access_token: JWT access token (ClickHouse Cloud feature).
      Should not be set if `username`/`password` are used.
    :param database:  The default database for the connection. If not set, ClickHouse Connect will use the
     default database for username.
    :param interface: Must be http or https.  Defaults to http, or to https if port is set to 8443 or 443
    :param port: The ClickHouse HTTP or HTTPS port. If not set will default to 8123, or to 8443 if secure=True
      or interface=https.
    :param secure: Use https/TLS. This overrides inferred values from the interface or port arguments.
    :param dsn: A string in standard DSN (Data Source Name) format. Other connection values (such as host or user)
      will be extracted from this string if not set otherwise.
    :param settings: ClickHouse server settings to be used with the session/every request
    :param generic_args: Used internally to parse DBAPI connection strings into keyword arguments and ClickHouse settings.
      It is not recommended to use this parameter externally.

    :param kwargs -- Recognized keyword arguments (used by the HTTP client), see below

    :param compress: Enable compression for ClickHouse HTTP inserts and query results.  True will select the preferred
      compression method (lz4).  A str of 'lz4', 'zstd', 'brotli', or 'gzip' can be used to use a specific compression type
    :param query_limit: Default LIMIT on returned rows.  0 means no limit
    :param connect_timeout:  Timeout in seconds for the http connection
    :param send_receive_timeout: Read timeout in seconds for http connection
    :param client_name: client_name prepended to the HTTP User Agent header. Set this to track client queries
      in the ClickHouse system.query_log.
    :param send_progress: Deprecated, has no effect.  Previous functionality is now automatically determined
    :param verify: Verify the server certificate in secure/https mode
    :param ca_cert: If verify is True, the file path to Certificate Authority root to validate ClickHouse server
     certificate, in .pem format.  Ignored if verify is False.  This is not necessary if the ClickHouse server
     certificate is trusted by the operating system.  To trust the maintained list of "global" public root
     certificates maintained by the Python 'certifi' package, set ca_cert to 'certifi'
    :param client_cert: File path to a TLS Client certificate in .pem format.  This file should contain any
      applicable intermediate certificates
    :param client_cert_key: File path to the private key for the Client Certificate.  Required if the private key
      is not included the Client Certificate key file
    :param session_id ClickHouse session id.  If not specified and the common setting 'autogenerate_session_id'
      is True, the client will generate a UUID1 session id
    :param pool_mgr Optional urllib3 PoolManager for this client.  Useful for creating separate connection
      pools for multiple client endpoints for applications with many clients
    :param http_proxy  http proxy address.  Equivalent to setting the HTTP_PROXY environment variable
    :param https_proxy https proxy address.  Equivalent to setting the HTTPS_PROXY environment variable
    :param server_host_name  This is the server host name that will be checked against a TLS certificate for
      validity.  This option can be used if using an ssh_tunnel or other indirect means to an ClickHouse server
      where the `host` argument refers to the tunnel or proxy and not the actual ClickHouse server
    :param tz_source Controls how the client determines the fallback timezone for DateTime columns without an
      explicit timezone. "auto" (default) auto-detects based on DST safety of server timezone. "server" always
      uses the server timezone. "local" always uses the local timezone.
    :param apply_server_timezone Deprecated. Use tz_source instead.
    :param tz_mode Controls timezone-aware behavior for UTC DateTime columns. "naive_utc" (default) returns
      naive UTC timestamps. "aware" forces timezone-aware UTC datetimes. "schema" returns datetimes that
      match the server's column definition which means timezone-aware when the column defines a timezone and naive
      for bare DateTime columns.
    :param utc_tz_aware Deprecated. Use tz_mode instead.
    :param autogenerate_session_id  If set, this will override the 'autogenerate_session_id' common setting.
    :param form_encode_query_params  If True, query parameters will be sent as form-encoded data in the request body
      instead of as URL parameters. This is useful for queries with large parameter sets that might exceed URL length
      limits. Only available for query operations (not inserts). Default: False
    :return: ClickHouse Connect Client instance
    """
    if dsn:
        parsed = urlparse(dsn)
        username = username or parsed.username
        password = password or parsed.password
        host = host or parsed.hostname
        port = port or parsed.port
        if parsed.path and (not database or database == '__default__'):
            database = parsed.path[1:].split('/')[0]
        database = database or parsed.path
        for k, v in parse_qs(parsed.query).items():
            kwargs[k] = v[0]
    use_tls = str(secure).lower() == 'true' or interface == 'https' or (not interface and str(port) in ('443', '8443'))
    if not host:
        host = 'localhost'
    if not interface:
        interface = 'https' if use_tls else 'http'
    port = port or default_port(interface, use_tls)
    if access_token and (username or password != ''):
        raise ProgrammingError('Cannot use both access_token and username/password')
    if username is None and 'user' in kwargs:
        username = kwargs.pop('user')
    if username is None and 'user_name' in kwargs:
        username = kwargs.pop('user_name')
    if password and username is None:
        username = 'default'
    if 'compression' in kwargs and 'compress' not in kwargs:
        kwargs['compress'] = kwargs.pop('compression')
    settings = settings or {}
    if interface.startswith('http'):
        if generic_args:
            client_params = signature(HttpClient).parameters
            for name, value in generic_args.items():
                if name in client_params:
                    kwargs[name] = value
                elif name == 'compression':
                    if 'compress' not in kwargs:
                        kwargs['compress'] = value
                else:
                    if name.startswith('ch_'):
                        name = name[3:]
                    settings[name] = value
        return HttpClient(interface, host, port, username, password, database, access_token,
                          settings=settings, **kwargs)
    raise ProgrammingError(f'Unrecognized client type {interface}')


def default_port(interface: str, secure: bool):
    if interface.startswith('http'):
        return 8443 if secure else 8123
    raise ValueError('Unrecognized ClickHouse interface')


async def create_async_client(*,
                              host: Optional[str] = None,
                              username: Optional[str] = None,
                              password: str = '',
                              database: str = '__default__',
                              interface: Optional[str] = None,
                              port: int = 0,
                              secure: Union[bool, str] = False,
                              dsn: Optional[str] = None,
                              settings: Optional[Dict[str, Any]] = None,
                              generic_args: Optional[Dict[str, Any]] = None,
                              executor_threads: int = 0,
                              executor: Union[ThreadPoolExecutor, None, DefaultThreadPoolExecutor] = NEW_THREAD_POOL_EXECUTOR,
                              **kwargs) -> AsyncClient:
    """
    The preferred method to get an async ClickHouse Connect Client instance.
    For sync version, see create_client.

    Unlike sync version, the 'autogenerate_session_id' setting by default is False.

    :param host: The hostname or IP address of the ClickHouse server. If not set, localhost will be used.
    :param username: The ClickHouse username. If not set, the default ClickHouse user will be used.
    :param password: The password for username.
    :param database:  The default database for the connection. If not set, ClickHouse Connect will use the
     default database for username.
    :param interface: Must be http or https.  Defaults to http, or to https if port is set to 8443 or 443
    :param port: The ClickHouse HTTP or HTTPS port. If not set will default to 8123, or to 8443 if secure=True
      or interface=https.
    :param secure: Use https/TLS. This overrides inferred values from the interface or port arguments.
    :param dsn: A string in standard DSN (Data Source Name) format. Other connection values (such as host or user)
      will be extracted from this string if not set otherwise.
    :param settings: ClickHouse server settings to be used with the session/every request
    :param generic_args: Used internally to parse DBAPI connection strings into keyword arguments and ClickHouse settings.
      It is not recommended to use this parameter externally
    :param executor_threads: 'max_worker' threads used by the client ThreadPoolExecutor.  If not set, the default
      of 4 + detected CPU cores will be used
    :param executor: Optional `ThreadPoolExecutor` to use for async operations.  If not set, a new `ThreadPoolExecutor`
      will be created with the number of threads specified by `executor_threads`.  If set to `None` it will use the
      default executor of the event loop.
    :param kwargs -- Recognized keyword arguments (used by the HTTP client), see below

    :param compress: Enable compression for ClickHouse HTTP inserts and query results.  True will select the preferred
      compression method (lz4).  A str of 'lz4', 'zstd', 'brotli', or 'gzip' can be used to use a specific compression type
    :param query_limit: Default LIMIT on returned rows.  0 means no limit
    :param connect_timeout:  Timeout in seconds for the http connection
    :param send_receive_timeout: Read timeout in seconds for http connection
    :param client_name: client_name prepended to the HTTP User Agent header. Set this to track client queries
      in the ClickHouse system.query_log.
    :param send_progress: Deprecated, has no effect.  Previous functionality is now automatically determined
    :param verify: Verify the server certificate in secure/https mode
    :param ca_cert: If verify is True, the file path to Certificate Authority root to validate ClickHouse server
     certificate, in .pem format.  Ignored if verify is False.  This is not necessary if the ClickHouse server
     certificate is trusted by the operating system.  To trust the maintained list of "global" public root
     certificates maintained by the Python 'certifi' package, set ca_cert to 'certifi'
    :param client_cert: File path to a TLS Client certificate in .pem format.  This file should contain any
      applicable intermediate certificates
    :param client_cert_key: File path to the private key for the Client Certificate.  Required if the private key
      is not included the Client Certificate key file
    :param session_id ClickHouse session id.  If not specified and the common setting 'autogenerate_session_id'
      is True, the client will generate a UUID1 session id
    :param pool_mgr Optional urllib3 PoolManager for this client.  Useful for creating separate connection
      pools for multiple client endpoints for applications with many clients
    :param http_proxy  http proxy address.  Equivalent to setting the HTTP_PROXY environment variable
    :param https_proxy https proxy address.  Equivalent to setting the HTTPS_PROXY environment variable
    :param server_host_name  This is the server host name that will be checked against a TLS certificate for
      validity.  This option can be used if using an ssh_tunnel or other indirect means to an ClickHouse server
      where the `host` argument refers to the tunnel or proxy and not the actual ClickHouse server
    :param tz_source Controls how the client determines the fallback timezone for DateTime columns without an
      explicit timezone. "auto" (default) auto-detects based on DST safety of server timezone. "server" always
      uses the server timezone. "local" always uses the local timezone.
    :param apply_server_timezone Deprecated. Use tz_source instead.
    :param tz_mode Controls timezone-aware behavior for UTC DateTime columns. "naive_utc" (default) returns
      naive UTC timestamps. "aware" forces timezone-aware UTC datetimes. "schema" returns datetimes that
      match the server's column definition which means timezone-aware when the column defines a timezone and naive
      for bare DateTime columns.
    :param utc_tz_aware Deprecated. Use tz_mode instead.
    :param autogenerate_session_id  If set, this will override the 'autogenerate_session_id' common setting.
    :param form_encode_query_params  If True, query parameters will be sent as form-encoded data in the request body
      instead of as URL parameters. This is useful for queries with large parameter sets that might exceed URL length
      limits. Only available for query operations (not inserts). Default: False
    :return: ClickHouse Connect Client instance
    """

    warnings.warn(
        "The current async client is a thread-pool wrapper around the sync client. "
        "A fully native async client is available for testing as a prerelease: "
        "pip install 'clickhouse-connect[async]==0.12.0rc1'. "
        "This prerelease branch is based on 0.11.0 and is gathering feedback ahead of 1.0.0, "
        "where it will become the default async implementation. It is a drop-in replacement "
        "with the same API surface. The main line includes additional updates that the native "
        "client will receive when merged into 1.0.0.",
        FutureWarning,
        stacklevel=2,
    )

    def _create_client():
        if 'autogenerate_session_id' not in kwargs:
            kwargs['autogenerate_session_id'] = False
        return create_client(host=host, username=username, password=password, database=database, interface=interface,
                             port=port, secure=secure, dsn=dsn, settings=settings, generic_args=generic_args, **kwargs)

    loop = asyncio.get_running_loop()
    _client = await loop.run_in_executor(None, _create_client)
    return AsyncClient(client=_client, executor_threads=executor_threads, executor=executor)



import struct

from clickhouse_connect.driver.types import ByteSource


class ByteArraySource(ByteSource):
    """
    ByteSource implementation for in-memory byte arrays.

    This class wraps a byte array and provides the ByteSource interface,
    allowing ClickHouse type decoders to read from in-memory data instead
    of a network stream.

    Used primarily for decoding variant-encoded values from JSON shared data
    where each value is a complete serialized type instance.
    """

    def __init__(self, data: bytes, encoding: str = "utf-8"):
        self.data = data
        self.pos = 0
        self.encoding = encoding

    def read_byte(self) -> int:
        if self.pos >= len(self.data):
            raise EOFError("Attempted to read past end of byte array")
        b = self.data[self.pos]
        self.pos += 1
        return b

    def read_bytes(self, sz: int) -> bytes:
        if self.pos + sz > len(self.data):
            raise EOFError(f"Attempted to read {sz} bytes, only {len(self.data) - self.pos} available")
        result = self.data[self.pos : self.pos + sz]
        self.pos += sz
        return result

    def read_leb128(self) -> int:
        sz = 0
        shift = 0
        while self.pos < len(self.data):
            b = self.read_byte()
            sz += (b & 0x7F) << shift
            if (b & 0x80) == 0:
                return sz
            shift += 7
        raise EOFError("Unexpected end while reading LEB128")

    def read_leb128_str(self) -> str:
        sz = self.read_leb128()
        return self.read_bytes(sz).decode(self.encoding)

    def read_uint64(self) -> int:
        return int.from_bytes(self.read_bytes(8), "little", signed=False)

    def read_int64(self) -> int:
        return int.from_bytes(self.read_bytes(8), "little", signed=True)

    def read_uint32(self) -> int:
        return int.from_bytes(self.read_bytes(4), "little", signed=False)

    def read_int32(self) -> int:
        return int.from_bytes(self.read_bytes(4), "little", signed=True)

    def read_uint16(self) -> int:
        return int.from_bytes(self.read_bytes(2), "little", signed=False)

    def read_int16(self) -> int:
        return int.from_bytes(self.read_bytes(2), "little", signed=True)

    def read_float32(self) -> float:
        return struct.unpack("<f", self.read_bytes(4))[0]

    def read_float64(self) -> float:
        return struct.unpack("<d", self.read_bytes(8))[0]

    # pylint: disable=too-many-return-statements
    def read_array(self, array_type: str, num_rows: int):  # type: ignore
        if array_type == "B":
            return [self.read_byte() for _ in range(num_rows)]
        if array_type == "H":
            return [self.read_uint16() for _ in range(num_rows)]
        if array_type == "I":
            return [self.read_uint32() for _ in range(num_rows)]
        if array_type == "Q":
            return [self.read_uint64() for _ in range(num_rows)]
        if array_type == "b":
            return [int.from_bytes([self.read_byte()], "little", signed=True) for _ in range(num_rows)]
        if array_type == "h":
            return [self.read_int16() for _ in range(num_rows)]
        if array_type == "i":
            return [self.read_int32() for _ in range(num_rows)]
        if array_type == "q":
            return [self.read_int64() for _ in range(num_rows)]
        if array_type == "f":
            return [self.read_float32() for _ in range(num_rows)]
        if array_type == "d":
            return [self.read_float64() for _ in range(num_rows)]
        raise NotImplementedError(f"Array type {array_type} not implemented for ByteArraySource")

    def read_str_col(self, num_rows, encoding, nullable=False, null_obj=None):  # type: ignore
        if num_rows != 1:
            raise NotImplementedError("read_str_col only supports num_rows=1 for single-value decoding")

        length = self.read_leb128()
        string_bytes = self.read_bytes(length)

        if encoding is None:
            return [string_bytes]

        return [string_bytes.decode(encoding)]

    def read_bytes_col(self, sz, num_rows):
        raise NotImplementedError("read_bytes_col not needed for single-value decoding")

    def read_fixed_str_col(self, sz, num_rows, encoding):
        raise NotImplementedError("read_fixed_str_col not needed for single-value decoding")

    def close(self):
        """No cleanup needed for byte arrays."""



import warnings

from clickhouse_connect.driver.exceptions import NotSupportedError

# Attributes resolved lazily by __getattr__ / _resolve_* functions:
#   np, pd, arrow, pl, pd_time_test, pd_extended_dtypes, PANDAS_VERSION, IS_PANDAS_2

# pylint: disable=import-outside-toplevel

_PANDAS_ATTRS = frozenset({"pd", "pd_time_test", "pd_extended_dtypes", "PANDAS_VERSION", "IS_PANDAS_2"})
_ALL_LAZY = frozenset({"np", "arrow", "pl"}) | _PANDAS_ATTRS


def _resolve_numpy():
    if "np" in globals():
        return
    try:
        import numpy

        globals()["np"] = numpy
    except ImportError:
        globals()["np"] = None


def _resolve_pandas():
    if "pd" in globals():
        return
    try:
        import pandas

        globals()["pd"] = pandas
        version = tuple(map(int, pandas.__version__.split(".")[:2]))
        globals()["PANDAS_VERSION"] = version
        is_v2 = version >= (2, 0)
        globals()["IS_PANDAS_2"] = is_v2
        globals()["pd_extended_dtypes"] = not pandas.__version__.startswith("0")
        if not is_v2:
            warnings.warn(
                "clickhouse-connect support for pandas 1.x is deprecated and will be removed in v1.0.0. "
                "Please upgrade to pandas 2.x or later.",
                DeprecationWarning,
                stacklevel=2,
            )
        try:
            from pandas.core.dtypes.common import (
                is_datetime64_dtype,
                is_timedelta64_dtype,
            )

            def combined_test(arr_or_dtype):
                return is_datetime64_dtype(arr_or_dtype) or is_timedelta64_dtype(arr_or_dtype)

            globals()["pd_time_test"] = combined_test
        except ImportError:
            try:
                from pandas.core.dtypes.common import is_datetime_or_timedelta_dtype

                globals()["pd_time_test"] = is_datetime_or_timedelta_dtype
            except ImportError as ex:
                raise NotSupportedError("pandas version does not contain expected test for temporal types") from ex
    except ImportError:
        globals()["pd"] = None
        globals()["PANDAS_VERSION"] = None
        globals()["IS_PANDAS_2"] = None
        globals()["pd_extended_dtypes"] = False
        globals()["pd_time_test"] = None


def _resolve_arrow():
    if "arrow" in globals():
        return
    try:
        import pyarrow

        globals()["arrow"] = pyarrow
    except ImportError:
        globals()["arrow"] = None


def _resolve_polars():
    if "pl" in globals():
        return
    try:
        import polars

        globals()["pl"] = polars
    except ImportError:
        globals()["pl"] = None


def __getattr__(name):
    if name in _PANDAS_ATTRS:
        _resolve_pandas()
    elif name == "np":
        _resolve_numpy()
    elif name == "arrow":
        _resolve_arrow()
    elif name == "pl":
        _resolve_polars()
    else:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
    return globals()[name]


def __dir__():
    return list(globals().keys()) + list(_ALL_LAZY - globals().keys())


# pylint: disable=redefined-outer-name
def check_numpy():
    _resolve_numpy()
    np = globals()["np"]
    if np:
        return np
    raise NotSupportedError("Numpy package is not installed")


def check_pandas():
    _resolve_pandas()
    pd = globals()["pd"]
    if pd:
        return pd
    raise NotSupportedError("Pandas package is not installed")


def check_arrow():
    _resolve_arrow()
    arrow = globals()["arrow"]
    if arrow:
        return arrow
    raise NotSupportedError("PyArrow package is not installed")


def check_polars():
    _resolve_polars()
    pl = globals()["pl"]
    if pl:
        return pl
    raise NotSupportedError("Polars package is not installed")



cdef class ResponseBuffer:
    cdef:
        unsigned long long buf_loc, buf_sz, slice_sz
        signed long long slice_start
        object gen, source, _exception_tag, open_marker, close_marker, carryover, exception_buf, last_message_data, current_chunk
        char* buffer
        char* slice
        unsigned char _read_byte_load(self) except ?255
        char* read_bytes_c(self, unsigned long long sz) except NULL
        Py_buffer buff_source
        cdef object _read_str_col(self, unsigned long long num_rows, char * encoding)
        cdef object _read_nullable_str_col(self, unsigned long long num_rows, char * encoding, object null_obj)
        cdef void _check_for_exception(self, object chunk) except *



from sqlalchemy.exc import CompileError
from sqlalchemy.sql import elements, sqltypes
from sqlalchemy.sql.compiler import SQLCompiler

from clickhouse_connect.cc_sqlalchemy import ArrayJoin
from clickhouse_connect.cc_sqlalchemy.datatypes.base import ChSqlaType
from clickhouse_connect.cc_sqlalchemy.sql import format_table


# pylint: disable=too-many-return-statements
def _resolve_ch_type_name(sqla_type):
    """Resolve a SQLAlchemy type instance to a ClickHouse type name string.

    Handles both native ChSqlaType instances which carry their ClickHouse name
    directly and generic SQLAlchemy types by mapping to reasonable ClickHouse defaults.
    """
    if isinstance(sqla_type, ChSqlaType):
        return sqla_type.name
    # Order matters so we need to check subtypes before parent types
    if isinstance(sqla_type, sqltypes.SmallInteger):
        return "Int16"
    if isinstance(sqla_type, sqltypes.BigInteger):
        return "Int64"
    if isinstance(sqla_type, sqltypes.Integer):
        return "Int32"
    if isinstance(sqla_type, sqltypes.Float):
        return "Float64"
    if isinstance(sqla_type, sqltypes.Numeric):
        p = sqla_type.precision or 18
        s = sqla_type.scale or 0
        return f"Decimal({p}, {s})"
    if isinstance(sqla_type, sqltypes.Boolean):
        return "Bool"
    if isinstance(sqla_type, sqltypes.DateTime):
        return "DateTime"
    if isinstance(sqla_type, sqltypes.Date):
        return "Date"
    if isinstance(sqla_type, sqltypes.String):
        return "String"
    return "String"


# pylint: disable=arguments-differ
class ChStatementCompiler(SQLCompiler):

    # pylint: disable=attribute-defined-outside-init,unused-argument
    def visit_delete(self, delete_stmt, visiting_cte=None, **kw):
        table = delete_stmt.table
        text = f"DELETE FROM {format_table(table)}"

        if delete_stmt.whereclause is not None:
            self._in_delete_where = True
            try:
                text += " WHERE " + self.process(delete_stmt.whereclause, **kw)
            finally:
                self._in_delete_where = False
        else:
            raise CompileError("ClickHouse DELETE statements require a WHERE clause. To delete all rows, use 'TRUNCATE TABLE' instead.")

        return text

    # pylint: disable=protected-access
    def visit_values(self, element, asfrom=False, from_linter=None, visiting_cte=None, **kw):
        """Compile a VALUES clause using ClickHouse's VALUES table function syntax.

        ClickHouse requires the column structure as the first argument:
            VALUES('col1 Type1, col2 Type2', (row1_val1, row1_val2), ...)

        This differs from standard SQL which places column names after the alias:
            (VALUES (row1), (row2)) AS name (col1, col2)

        Compatible with both SQLAlchemy 1.4 and 2.x.
        """
        if getattr(element, "_independent_ctes", None):
            self._dispatch_independent_ctes(element, kw)

        structure = ", ".join(
            f"{col.name} {_resolve_ch_type_name(col.type)}"
            for col in element.columns
        )

        kw.setdefault("literal_binds", element.literal_binds)
        tuples = ", ".join(
            self.process(
                elements.Tuple(types=element._column_types, *elem).self_group(),
                **kw,
            )
            for chunk in element._data
            for elem in chunk
        )

        structure_literal = self.render_literal_value(structure, sqltypes.String())
        v = f"VALUES({structure_literal}, {tuples})"

        # SA 2.x has _unnamed; SA 1.4 uses name=None for unnamed values
        is_unnamed = getattr(element, "_unnamed", element.name is None)
        if is_unnamed:
            name = None
        elif isinstance(element.name, elements._truncated_label):
            name = self._truncated_identifier("values", element.name)
        else:
            name = element.name

        lateral = "LATERAL " if element._is_lateral else ""

        if asfrom:
            if from_linter:
                # SA 2.x has _de_clone(); SA 1.4 doesn't
                key = element._de_clone() if hasattr(element, "_de_clone") else element
                from_linter.froms[key] = (
                    name if name is not None else "(unnamed VALUES element)"
                )

            if visiting_cte is not None and visiting_cte.element is element:
                if element._is_lateral:
                    raise CompileError(
                        "Can't use a LATERAL VALUES expression inside of a CTE"
                    )
                v = f"SELECT * FROM {v}"
            elif name:
                kw["include_table"] = False
                v = f"{lateral}{v}{self.get_render_as_alias_suffix(self.preparer.quote(name))}"
            else:
                v = f"{lateral}{v}"

        return v

    def visit_array_join(self, array_join_clause, asfrom=False, from_linter=None, **kw):
        left = self.process(array_join_clause.left, asfrom=True, from_linter=from_linter, **kw)
        join_type = "LEFT ARRAY JOIN" if array_join_clause.is_left else "ARRAY JOIN"

        parts = []
        for col, alias in array_join_clause.array_columns:
            col_text = self.process(col, **kw)
            if alias is not None:
                col_text += f" AS {self.preparer.quote(alias)}"
            parts.append(col_text)

        return f"{left} {join_type} {', '.join(parts)}"

    def visit_join(self, join, **kw):
        if isinstance(join, ArrayJoin):
            return self.visit_array_join(join, **kw)

        left = self.process(join.left, **kw)
        right = self.process(join.right, **kw)
        onclause = join.onclause

        is_cross = getattr(join, "_is_cross", False) or onclause is None
        if getattr(join, "full", False):
            join_type = "FULL OUTER JOIN"
        elif is_cross:
            join_type = "CROSS JOIN"
        elif join.isouter:
            join_type = "LEFT OUTER JOIN"
        else:
            join_type = "INNER JOIN"

        # ClickHouse modifiers: [GLOBAL] [ALL|ANY|ASOF] <join_type>
        distribution = getattr(join, "distribution", None)
        strictness = getattr(join, "strictness", None)
        parts = []
        if distribution:
            parts.append(distribution)
        if strictness:
            parts.append(strictness)
        parts.append(join_type)
        join_kw = " ".join(parts)

        text = f"{left} {join_kw} {right}"

        using_columns = getattr(join, "using_columns", None)
        if using_columns:
            # Process the onclause so the from-linter registers the
            # table relationship, but render USING syntax instead.
            if onclause is not None:
                self.process(onclause, **kw)
            quoted = ", ".join(self.preparer.quote(col) for col in using_columns)
            text += f" USING ({quoted})"
        elif not is_cross and onclause is not None:
            text += " ON " + self.process(onclause, **kw)

        return text

    def visit_column(self, column, add_to_result_map=None, include_table=True, result_map_targets=(), ambiguous_table_name_map=None, **kw):
        if getattr(self, "_in_delete_where", False):
            return self.preparer.quote(column.name)

        return super().visit_column(
            column,
            add_to_result_map=add_to_result_map,
            include_table=include_table,
            result_map_targets=result_map_targets,
            **kw,
        )

    # Abstract methods required by SQLCompiler
    def delete_extra_from_clause(self, delete_stmt, from_table, extra_froms, from_hints, **kw):
        raise NotImplementedError("ClickHouse doesn't support DELETE with extra FROM clause")

    def update_from_clause(self, update_stmt, from_table, extra_froms, from_hints, **kw):
        raise NotImplementedError("ClickHouse doesn't support UPDATE with FROM clause")

    # pylint: disable=unused-argument
    def visit_empty_set_expr(self, element_types, **kw):
        return "SELECT 1 WHERE 1=0"

    def visit_sequence(self, sequence, **kw):
        raise NotImplementedError("ClickHouse doesn't support sequences")

    def group_by_clause(self, select, **kw):
        """Render GROUP BY using label aliases instead of full expressions."""
        kw["_ch_group_by"] = True
        return super().group_by_clause(select, **kw)

    # pylint: disable=protected-access
    def visit_label(
        self,
        label,
        within_columns_clause=False,
        render_label_as_label=None,
        **kw,
    ):
        ch_group_by = kw.pop("_ch_group_by", False)
        if ch_group_by and not within_columns_clause and render_label_as_label is None:
            if isinstance(label.name, elements._truncated_label):
                labelname = self._truncated_identifier("colident", label.name)
            else:
                labelname = label.name
            return self.preparer.format_label(label, labelname)
        return super().visit_label(
            label,
            within_columns_clause=within_columns_clause,
            render_label_as_label=render_label_as_label,
            **kw,
        )

    # pylint: disable=protected-access
    def _compose_select_body(self, text, select, compile_state, inner_columns, froms, byfrom, toplevel, kwargs):
        ch_final = getattr(select, "_ch_final", set())
        ch_sample = getattr(select, "_ch_sample", {})

        if ch_final or ch_sample:
            mods = {}
            for target in ch_final | set(ch_sample):
                parts = []
                if target in ch_final:
                    parts.append("FINAL")
                if target in ch_sample:
                    parts.append(f"SAMPLE {ch_sample[target]}")
                mods[target] = " ".join(parts)

            prev = getattr(self, "_ch_from_modifiers", None)
            self._ch_from_modifiers = mods
            try:
                return super()._compose_select_body(text, select, compile_state, inner_columns, froms, byfrom, toplevel, kwargs)
            finally:
                self._ch_from_modifiers = prev

        return super()._compose_select_body(text, select, compile_state, inner_columns, froms, byfrom, toplevel, kwargs)

    def visit_table(self, table, asfrom=False, iscrud=False, ashint=False, fromhints=None, enclosing_alias=None, **kwargs):
        result = super().visit_table(
            table, asfrom=asfrom, iscrud=iscrud, ashint=ashint, fromhints=fromhints, enclosing_alias=enclosing_alias, **kwargs
        )
        if asfrom and enclosing_alias is None:
            mods = getattr(self, "_ch_from_modifiers", None)
            if mods and table in mods:
                result += " " + mods[table]
        return result

    def visit_alias(self, alias, asfrom=False, **kwargs):
        result = super().visit_alias(alias, asfrom=asfrom, **kwargs)
        if asfrom:
            mods = getattr(self, "_ch_from_modifiers", None)
            if mods and alias in mods:
                result += " " + mods[alias]
        return result



import logging
import json as py_json
from collections import OrderedDict
from typing import Any

try:
    import orjson
    any_to_json = orjson.dumps  # pylint: disable=no-member
except ImportError:
    orjson = None

try:
    import ujson

    def _ujson_to_json(obj: Any) -> bytes:
        return ujson.dumps(obj).encode()  # pylint: disable=c-extension-no-member
except ImportError:
    ujson = None
    _ujson_to_json = None


def _pyjson_to_json(obj: Any) -> bytes:
    return py_json.dumps(obj, separators=(',', ':')).encode()


logger = logging.getLogger(__name__)
_to_json = OrderedDict()
_to_json['orjson'] = orjson.dumps if orjson else None  # pylint: disable=no-member
_to_json['ujson'] = _ujson_to_json if ujson else None
_to_json['python'] = _pyjson_to_json

any_to_json = _pyjson_to_json


def set_json_library(impl: str = None):
    global any_to_json # pylint: disable=global-statement
    if impl:
        func = _to_json.get(impl)
        if func:
            any_to_json = func
            return
        raise NotImplementedError(f'JSON library {impl} is not supported')
    for library, func in _to_json.items():
        if func:
            logger.debug('Using %s library for writing JSON byte strings', library)
            any_to_json = func
            break


set_json_library()



import sys
import warnings

from clickhouse_connect.driver import create_client, create_async_client


if sys.version_info < (3, 10):
    warnings.warn(
        "Python 3.9 support is deprecated and will be removed in a future release. "
        "This version of clickhouse-connect may stop working with Python 3.9 unexpectedly.",
        DeprecationWarning,
        stacklevel=2
    )


driver_name = 'clickhousedb'

get_client = create_client
get_async_client = create_async_client



from typing import Union

from clickhouse_connect.dbapi.cursor import Cursor
from clickhouse_connect.driver import create_client
from clickhouse_connect.driver.query import QueryResult


class Connection:
    """
    See :ref:`https://peps.python.org/pep-0249/`
    """
    # pylint: disable=too-many-arguments
    def __init__(self,
                 dsn: str = None,
                 username: str = '',
                 password: str = '',
                 host: str = None,
                 database: str = None,
                 interface: str = None,
                 port: int = 0,
                 secure: Union[bool, str] = False,
                 **kwargs):
        self.client = create_client(host=host,
                                    username=username,
                                    password=password,
                                    database=database,
                                    interface=interface,
                                    port=port,
                                    secure=secure,
                                    dsn=dsn,
                                    generic_args=kwargs)

        self.client._add_integration_tag("sqlalchemy")
        self.timezone = self.client.server_tz

    def close(self):
        self.client.close()

    def commit(self):
        pass

    def rollback(self):
        pass

    def command(self, cmd: str):
        return self.client.command(cmd)

    def raw_query(self, query: str) -> QueryResult:
        return self.client.query(query)

    def cursor(self):
        return Cursor(self.client)



import cython

import numpy as np

from .buffer cimport ResponseBuffer

@cython.boundscheck(False)
@cython.wraparound(False)
def read_numpy_array(buffer, np_type: str, unsigned long long num_rows):
    dtype = np.dtype(np_type)
    cdef sz = dtype.itemsize * num_rows
    cdef char * source
    if isinstance(buffer, ResponseBuffer):
        source = (<ResponseBuffer>buffer).read_bytes_c(dtype.itemsize * num_rows)
        return np.frombuffer(source[:sz], dtype, num_rows)
    return np.frombuffer(buffer.read_bytes(sz), dtype, num_rows)



import array

from datetime import datetime, date, tzinfo
from ipaddress import IPv4Address
from typing import Sequence, Optional, Any
from uuid import UUID, SafeUUID

from clickhouse_connect.driver import tzutil
from clickhouse_connect.driver.common import int_size
from clickhouse_connect.driver.errors import NONE_IN_NULLABLE_COLUMN
from clickhouse_connect.driver.types import ByteSource
from clickhouse_connect.driver import options


MONTH_DAYS = (0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365)
MONTH_DAYS_LEAP = (0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366)


def read_ipv4_col(source: ByteSource, num_rows: int):
    column = source.read_array('I', num_rows)
    fast_ip_v4 = IPv4Address.__new__
    new_col = []
    app = new_col.append
    for x in column:
        ipv4 = fast_ip_v4(IPv4Address)
        ipv4._ip = x  # pylint: disable=protected-access
        app(ipv4)
    return new_col


def read_datetime_col(source: ByteSource, num_rows: int, tz_info: Optional[tzinfo]):
    src_array = source.read_array('I', num_rows)
    if tz_info is None:
        fts = tzutil.utcfromtimestamp
        return [fts(ts) for ts in src_array]
    fts = datetime.fromtimestamp
    return [fts(ts, tz_info) for ts in src_array]


def epoch_days_to_date(days: int) -> date:
    cycles400, rem = divmod(days + 134774, 146097)
    cycles100, rem = divmod(rem, 36524)
    cycles, rem = divmod(rem, 1461)
    years, rem = divmod(rem, 365)
    year = (cycles << 2) + cycles400 * 400 + cycles100 * 100 + years + 1601
    if years == 4 or cycles100 == 4:
        return date(year - 1, 12, 31)
    m_list = MONTH_DAYS_LEAP if years == 3 and (year == 2000 or year % 100 != 0) else MONTH_DAYS
    month = (rem + 24) >> 5
    while rem < m_list[month]:
        month -= 1
    return date(year, month + 1, rem + 1 - m_list[month])


def read_date_col(source: ByteSource, num_rows: int):
    column = source.read_array('H', num_rows)
    return [epoch_days_to_date(x) for x in column]


def read_date32_col(source: ByteSource, num_rows: int):
    column = source.read_array('l' if int_size == 2 else 'i', num_rows)
    return [epoch_days_to_date(x) for x in column]


def read_uuid_col(source: ByteSource, num_rows: int):
    v = source.read_array('Q', num_rows * 2)
    empty_uuid = UUID(int=0)
    new_uuid = UUID.__new__
    unsafe = SafeUUID.unsafe
    oset = object.__setattr__
    column = []
    app = column.append
    for i in range(num_rows):
        ix = i << 1
        int_value = v[ix] << 64 | v[ix + 1]
        if int_value == 0:
            app(empty_uuid)
        else:
            fast_uuid = new_uuid(UUID)
            oset(fast_uuid, 'int', int_value)
            oset(fast_uuid, 'is_safe', unsafe)
            app(fast_uuid)
    return column


def read_nullable_array(source: ByteSource, array_type: str, num_rows: int, null_obj: Any):
    null_map = source.read_bytes(num_rows)
    column = source.read_array(array_type, num_rows)
    return [null_obj if null_map[ix] else column[ix] for ix in range(num_rows)]


def build_nullable_column(source: Sequence, null_map: bytes, null_obj: Any):
    return [source[ix] if null_map[ix] == 0 else null_obj for ix in range(len(source))]


def build_lc_nullable_column(index: Sequence, keys: array.array, null_obj: Any):
    column = []
    for key in keys:
        if key == 0:
            column.append(null_obj)
        else:
            column.append(index[key])
    return column


def to_numpy_array(column: Sequence):
    np = options.np
    arr = np.empty((len(column),), dtype=np.object)
    arr[:] = column
    return arr


def pivot(data: Sequence[Sequence], start_row: int, end_row: int) -> Sequence[Sequence]:
    return tuple(zip(*data[start_row: end_row]))


def write_str_col(column: Sequence, nullable: bool, encoding: Optional[str], dest: bytearray) -> int:
    app = dest.append
    for x in column:
        if not x:
            if not nullable and x is None:
                return NONE_IN_NULLABLE_COLUMN
            app(0)
        else:
            if encoding:
                x = x.encode(encoding)
            else:
                x = bytes(x)
            sz = len(x)
            while True:
                b = sz & 0x7f
                sz >>= 7
                if sz == 0:
                    app(b)
                    break
                app(0x80 | b)
            dest += x
    return 0
